/*
 * main.cpp
 *
 *  Created on: 19.03.2018
 *      Author: hva
 */

#include <sched.h>

#include <abmt/task.h>
#include <unistd.h>

#include <functional>
#include <memory>
#include <map>
#include <iostream>
#include <fstream>
#include <filesystem>
#include <clocale>
#include <time.h>

#include <abmt/util/bt_exception.h>


#include <abmt/rt/com.h>
#include <abmt/rt/raster.h>
#include <abmt/serialize_utils.h>
#include <abmt/mutex.h>

#include <abmt/json.h>
#include <abmt/time.h>

using namespace std;
using json = abmt::json;

bool global_stop = false;

void on_sig(int sig){
	global_stop = true;
}

abmt::time abmt::time::now(){
	timespec t;
	clock_gettime(CLOCK_REALTIME, &t);
	int64_t res;
	res = t.tv_sec * 1000LL*1000*1000 + t.tv_nsec; // 1000LL very important for 32bit systems
	return time(res);
}

struct daq_element{
	uint32_t signal_index;
	uint32_t daq_index;
};

using daq_list_t = std::vector<daq_element>;

class model_adatper_live: public abmt::rt::model_adatper{
public:

	std::vector<daq_list_t> daq_lists;
	std::vector<uint32_t>   paq_list;
	std::function<void()> on_version_error = []{};

	void set_model(abmt::rt::model* m){
		abmt::rt::model_adatper::set_model(m);
		daq_lists.clear();
		daq_lists.resize(m->rasters.length);
	}


	virtual void on_hello(uint16_t version){
		if(version != ABMT_VERSION){
			on_version_error();
		}
	};

	virtual void on_request_signal_def(){
		size_t siglal_def_size = 4; // 4 = uint32_t for number of signals;
		for(auto s: mdl->signals){
			auto sz = s.type.get_def_size();
			siglal_def_size += sz;
			auto name_len = strlen(s.name);
			if(name_len > 255){
				name_len = 255;
			}
			siglal_def_size += name_len + 1; // 1 for len_byte
		}

		auto blk = abmt::blob_shared(siglal_def_size);
		blk.set((uint32_t) mdl->signals.length, 0);
		char* ptr = (char*)blk.data;
		size_t pos = 4;
		ptr += 4;
		for(auto s: mdl->signals){
			auto name_len = strlen(s.name);
			if(name_len > 255){
				name_len = 255;
			}
			*(uint8_t*)ptr = (uint8_t)name_len;
			++ptr; ++pos;
			memcpy(ptr,s.name,name_len);
			pos += name_len;
			ptr += name_len;
			size_t def_size = s.type.get_def(ptr,blk.size - pos);
			pos += def_size;
			ptr += def_size;
		}
		send(abmt::rt::cmd::ack_signal_def,blk);
	};

	virtual void on_request_parameter_def(){
		size_t siglal_def_size = 4; // 4 = uint32_t for number of parameters;
		for(auto p: mdl->parameters){
			auto sz = p.parameter->get_type().get_def_size();
			siglal_def_size += sz;
			auto name_len = strlen(p.name);
			if(name_len > 255){
				name_len = 255;
			}
			siglal_def_size += name_len + 1; // 1 for len_byte
		}

		auto blk = abmt::blob_shared(siglal_def_size);
		blk.set((uint32_t) mdl->parameters.length, 0);
		char* ptr = (char*)blk.data;
		size_t pos = 4;
		ptr += 4;
		for(auto p: mdl->parameters){
			auto name_len = strlen(p.name);
			if(name_len > 255){
				name_len = 255;
			}
			*(uint8_t*)ptr = (uint8_t)name_len;
			++ptr; ++pos;
			memcpy(ptr,p.name,name_len);
			pos += name_len;
			ptr += name_len;
			size_t def_size = p.parameter->get_type().get_def(ptr,blk.size - pos);
			pos += def_size;
			ptr += def_size;
		}
		send(abmt::rt::cmd::ack_parameter_def,blk);
	};

	virtual void on_set_daq_list(abmt::blob& lst){
		for(auto& lst:daq_lists){
			lst.clear();
		}

		auto num_elements = lst.size/sizeof(uint32_t);
		for(uint32_t i = 0; i < num_elements; ++i){
			auto idx = lst.get<uint32_t>(i*sizeof(uint32_t));
			if(idx < mdl->signals.length){
				auto& signal = mdl->signals[idx];
				daq_lists[signal.raster_index].push_back({idx,i});
			}
		}

		send(abmt::rt::cmd::ack_set_daq_list);

	};

	virtual void on_set_paq_list(abmt::blob& lst){
		paq_list.clear();

		auto num_elements = lst.size/sizeof(uint32_t);
		for(uint32_t i = 0; i < num_elements; ++i){
			auto idx = lst.get<uint32_t>(i*sizeof(uint32_t));
			if(idx < mdl->parameters.length){
				paq_list.push_back(idx);
			}
		}
		send(abmt::rt::cmd::ack_set_paq_list);

		if(paq_list.size() > 0){
			send_all_parameters();
		}
	};

	virtual void on_set_parameter(abmt::blob& data){
		if(data.size < 8){
			return;
		}
		auto idx = data.get<uint32_t>(0);
		//auto len = data.as<uint32_t>(4);
		if(mdl->parameters.length > idx){
			auto param =  mdl->parameters[idx].parameter;

			param->lock();
			// write data to parameter
			auto exp =param->get_type();
			auto data_blk = data.sub_blob(8);
			exp.deserialize(data_blk.data,data_blk.size);

			// send parameter
			uint32_t size = exp.get_size();
			auto blk = abmt::blob_shared(size + sizeof(abmt::rt::paq_data_hdr));

			abmt::rt::paq_data_hdr hdr;
			hdr.size = size;
			auto data_area = blk.sub_blob(sizeof(hdr),size);
			exp.serialize(data_area.data,data_area.size);

			param->unlock();

			// find list_idx
			bool found = false;
			for(uint32_t i = 0; i < paq_list.size(); ++i){
				if(paq_list[i] == idx){
					hdr.list_idx = i;
					found = true;
					break;
				}
			}
			blk.set(hdr, 0);
			if(found){
				send(abmt::rt::cmd::paq_data,blk);
			}
			save_parameters();
		}else{
			cout << "on_set_parameter: index out of range!" << endl;
		}
	};

	virtual void on_command(string cmd){
		if(cmd == "quit"){
			global_stop = true;
		}
	}

	void save_parameters(){
		json res;
		for(auto p: mdl->parameters){
			std::string key = p.name;
			res[p.name] = abmt::serialize::to_json(p.parameter->get_type());
		}
		ofstream ofs("parameters.json");
		ofs << res.dump();
		ofs.close();
	}

	virtual ~model_adatper_live(){

	};

	void send_data(size_t raster){
		auto now = abmt::time::now();
		auto& daq_list = daq_lists[raster];
		if(daq_list.size() == 0){
			return;
		}

		size_t size_to_send = 0;
		for(auto& d:daq_list){
			size_to_send += sizeof(abmt::rt::daq_data_hdr) + mdl->signals[d.signal_index].type.get_size();
		}

		if(size_to_send > 100*1024*1024){
			cout << "Daq-Data over 100M (raster: " << raster << ")" << endl;
		}
		auto blk = abmt::blob_shared(size_to_send);

		size_t mem_pos = 0;
		for(auto& d:daq_list){
			uint32_t signal_size =  mdl->signals[d.signal_index].type.get_size();

			abmt::rt::daq_data_hdr hdr = {d.daq_index, signal_size, now.ns_since_1970};
			blk.set(hdr, mem_pos);
			mem_pos += sizeof(hdr);

			auto data_area = blk.sub_blob(mem_pos,signal_size);
			mdl->signals[d.signal_index].type.serialize(data_area.data,data_area.size);
			mem_pos += signal_size;
		}

		send(abmt::rt::cmd::daq_data,blk);
	}

	void send_all_parameters(){
		size_t size_to_send = 0;
		for(auto& idx: paq_list){
			auto param = mdl->parameters[idx].parameter;
			param->lock();
			size_to_send += sizeof(abmt::rt::paq_data_hdr) + param->get_type().get_size();
			param->unlock();
		}


		auto blk = abmt::blob_shared(size_to_send);
		size_t mem_pos = 0;

		uint32_t entry_num = 0;
		for(auto& idx: paq_list){
			auto param = mdl->parameters[idx].parameter;
			param->lock();

			auto exp = mdl->parameters[idx].parameter->get_type();
			uint32_t size = exp.get_size();

			abmt::rt::paq_data_hdr hdr = {entry_num, size};
			blk.set(hdr, mem_pos);
			mem_pos += sizeof(hdr);

			auto data_area = blk.sub_blob(mem_pos,size);
			exp.serialize(data_area.data,data_area.size);
			mem_pos += size;

			param->unlock();
			entry_num++;
		}

		send(abmt::rt::cmd::paq_data,blk);
	}

};

class runtime_task: public abmt::task{
public:
	bool stop = false;

	std::function<void(void)> pre_tick  = []{};
	std::function<void(void)> post_tick = []{};

	runtime_task(std::function<void()> tsk, bool start_now = true):task(tsk,start_now){

	}
};

class timed_task: public runtime_task{
public:
	abmt::rt::raster_sync* r;

	abmt::time interval;
	abmt::time start_time;
	abmt::time min_recover_delay;
	uint64_t n_ticks = 0;

	timed_task(abmt::rt::raster_sync* r, abmt::time interval):runtime_task([this]{this->run();},false), r(r), interval(interval){
		min_recover_delay = interval/2;
	}

	void run(){
		try{
			usleep(10*1000); // give async tasks some time to init (see example/io/udp.acfg)
			start_time = abmt::time::now();
			pre_tick();
			r->init_tick();
			post_tick();
			do{
				++n_ticks;
				auto t_next_tick = start_time + n_ticks*interval;
				abmt::time delay =  t_next_tick - abmt::time::now();

				if(delay > min_recover_delay){
					usleep(delay.us());
				}else{
					usleep(min_recover_delay.us());
				}
				pre_tick();
				r->tick();
				post_tick();

			}while(stop == false && global_stop == false);
			r->final();
		}catch(abmt::util::bt_exception& b){
			cout << "An error occured in task " << r->name << ": " << b.what() << endl;
			b.print_backtrace();
			global_stop = true;
		}
	}

};


class poll_task: public runtime_task{
public:
	abmt::rt::raster_async* r;

	poll_task(abmt::rt::raster_async* r):runtime_task([this]{this->run();},false), r(r){

	}

	void run(){
		try{
			r->init();
			usleep(10*1000); // give async tasks some time to init (see example/io/udp.acfg)
			while(stop == false && global_stop == false){
				if(r->poll() == false){
					usleep(100);
					continue;
				}
				pre_tick();
				r->init_tick();
				post_tick();
				break;
			}
			while(stop == false && global_stop == false){
				if(r->poll() == false){
					usleep(100);
					continue;
				}
				pre_tick();
				r->tick();
				post_tick();
			}
			r->final();
		}catch(abmt::util::bt_exception& b){
			cout << "An error occured in task " << r->name << ": " << b.what() << endl;
			b.print_backtrace();
			global_stop = true;
		}
	}

};

class connection{
public:
	virtual void   write(void* data, size_t size) = 0;
	virtual size_t read(void* data, size_t size) = 0;
	virtual ~connection(){};
};
using connection_ptr = std::shared_ptr<connection>;


class com{
public:
	connection_ptr con;
	std::string sys_info;
};


class model_com_live{
public:
	abmt::io::event_list& e;
	abmt::io::timer timer;
	abmt::io::tcp_ptr tcp_con;

	bool connected = false;
	model_adatper_live adapter;
	abmt::rt::model* mdl = 0;

	model_com_live(abmt::io::event_list& lst):e(lst),timer(e,1000){
		connect();
		timer.on_expiraton = [this]{
			if(connected == true){
				adapter.send_model_online();
			}else{
				connect();
			}
		};
		adapter.on_version_error = [this]{
			static int output_counter = 0;
			if(output_counter == 0){
				cout << "Error connecting to ABMT. Wrong version. Expected version " << ABMT_VERSION << endl;
				output_counter = 20;
			}else{
				output_counter--;
			}
			tcp_con->close();
			tcp_con->on_open = []{}; // temporary butfix becaue on_open is delayed...
		};
	}

	void connect(){
		connected = false;
		tcp_con = abmt::io::tcp::connect_to(e,{"127.0.0.1",15101});
		tcp_con->on_open = [this]{
			connected = true;
		};
		adapter.set_connection(tcp_con);
		// override on_close set by set_connection
		tcp_con->on_close = [this]{
			connected = false;
			adapter.connection.set_sink(abmt::io::sink());
			adapter.connection.set_source(abmt::io::source());
		};



	}

	void set_model(abmt::rt::model* m){
		mdl = m;
		adapter.set_model(mdl);
	}
};


int main(){

	/*
	cpu_set_t  mask;
	CPU_ZERO(&mask);
	CPU_SET(2, &mask);
	int result = sched_setaffinity(0, sizeof(mask), &mask);
	if(result == -1){
		std::cout << "unable to set sched_setaffinit for process. This could cause realtime-issues. Error-message: "<< std::strerror(errno) << std::endl;
	}
 	*/

	signal(SIGHUP,on_sig);
	signal(SIGINT,on_sig);
	signal(SIGQUIT,on_sig);

	std::setlocale(LC_ALL, "C.UTF-8");

	auto model_ptr = abmt::rt::get_model();

	abmt::io::event_list e;
	model_com_live m(e);
	m.adapter.set_model(model_ptr);

	if(std::filesystem::exists("parameters.json")){
		ifstream f("parameters.json");
		stringstream ss;
		ss << f.rdbuf();
		f.close();
		json params= json::parse(ss.str());
		for(auto p: params){
			for(auto mdl_param: model_ptr->parameters){
				std::string name(mdl_param.name);
				if(name == p.key){
					auto exp = mdl_param.parameter->get_type();
					try{
						abmt::serialize::from_json(p.value,exp);
					}catch(...){
						std::cout << "error parsing parameter: " << name << std::endl;
					}
					break;
				}
			}
		}
	}

	std::map<std::string, std::shared_ptr<runtime_task>> timed_tasks;

	for(size_t i = 0; i < model_ptr->rasters.length; ++i){
		auto r = model_ptr->rasters[i];

		abmt::rt::raster_sync* r_sync = dynamic_cast<abmt::rt::raster_sync*>(r);
		if(r_sync){
			auto task = make_shared<timed_task>(r_sync,abmt::time::ns(r_sync->interval_ns));
			task->post_tick = [&m,i]{
				m.adapter.send_data(i);
			};
			timed_tasks[r_sync->name] = task;
			task->start();
		}

		abmt::rt::raster_async* r_async = dynamic_cast<abmt::rt::raster_async*>(r);
		if(r_async){
			auto task = make_shared<poll_task>(r_async);
			task->post_tick = [&m,i]{
				m.adapter.send_data(i);
			};
			timed_tasks[r_async->name] = task;
			task->start();
		}
	}

	try{
		while(global_stop == false){
			e.wait(100);
		}
		// quit
		bool all_tasks_ready = false;
		bool force_quit = false;
		abmt::io::timer quit_timer(e, 1000, [&](){ force_quit = true; });
		while(all_tasks_ready == false && force_quit == false){
			all_tasks_ready = true;
			for(auto t: timed_tasks){
				if(t.second->is_running()){
					all_tasks_ready = false;
					break;
				}
			}
			e.wait(100);
		}

		if(force_quit){
			cout << "Force exit. The following rasters are still running:" << endl;
			for(auto t: timed_tasks){
				if(t.second->is_running()){
					cout << t.first << endl;
				}
			}
			return 1;
		}

	}catch(abmt::util::bt_exception& b){
		cout << "An error occured in the main task: " << b.what() << endl;
		b.print_backtrace();
		global_stop = true;
	}
	return 0;
}
