/**
 * Author: Hendrik van Arragon, 2021
 * SPDX-License-Identifier: MIT
 */

#ifndef SHARED_ABMT_RT_COM_H_
#define SHARED_ABMT_RT_COM_H_

#include <functional>
#include <cstdint>
#include <abmt/blob.h>
#include <abmt/io/protocols/s2p.h>
#include <abmt/rt/model.h>
#include <abmt/version.h>
#include <abmt/io/eio.h>


namespace abmt{
namespace rt{

/**
 * Concept
 * =======
 * The communication between abmt and a model-instance is done via adapters.
 * The modelside is handled with the model_adapter. The observer via the observer_adapter.
 * Both base-adapters map the incoming requests to their virtual interface-functions.
 * This abstraction is done to allow different dataacquisition-implementations.
 * Embedded systems have special needs here.
 *
 * Communication
 * -------------
 * Model- and observer-side receive the others side events with an on_-prefix. Sending requests
 * can be done with a send_-prefix method.
 *
 */


enum class cmd: uint8_t{
	// observer to model (requests)
	hello = 0x00,
	request_signal_def,
	request_parameter_def,
	set_daq_list,           // set list and list_number
	set_paq_list,
	set_prameter,
	command,


	// model to observer (responses)
	model_online = 0x10,
	ack_signal_def,
	ack_parameter_def,
	ack_set_daq_list,      // returns sizes and list_number
	ack_set_paq_list,


	// continuous messages to observer
	daq_data  = 0x20,
	paq_data,

};

#pragma pack(push, 1)

struct daq_data_hdr{
	uint32_t list_idx;
	uint32_t size;
	int64_t  time;
};

struct paq_data_hdr{
	uint32_t list_idx;
	uint32_t size;
};

#pragma pack(pop)

// Helper:
// for i in $(cat model_com.h | grep -e ",[\t ]*\$" | awk '{print $1}'); do echo "std::function<void()> on_$i = []{};"; done

class observer_adapter{
public:

	/**
	 * Interface:
	 */

	abmt::io::s2p s2p_connection;

	virtual void on_model_online(){

	};

	virtual void on_ack_signal_def(abmt::blob& p){

	};

	virtual void on_ack_parameter_def(abmt::blob& p){

	};

	virtual void on_ack_set_daq_list(){

	};

	virtual void on_ack_set_paq_list(){

	};

	virtual void on_ack_set_data(){

	};

	virtual void on_daq_data(abmt::blob& blk){

	};

	virtual void on_paq_data(abmt::blob& blk){

	};

	virtual void send_hello(){
		uint16_t version = ABMT_VERSION;
		send(cmd::hello, &version, sizeof(version));
	}

	virtual void send_request_signal_def(){
		send(cmd::request_signal_def,0,0);
	}

	virtual void send_request_parameter_def(){
		send(cmd::request_parameter_def,0,0);
	}

	void send_daq_list(abmt::blob& lst){
		send(cmd::set_daq_list,lst.data,lst.size);
	}

	void send_paq_list(abmt::blob& lst){
		send(cmd::set_paq_list,lst.data,lst.size);
	}

	void send_parameter(abmt::blob& data){
		send(cmd::set_prameter,data.data,data.size);
	}

	void send_command(std::string command){
		send(cmd::command,command.c_str(),command.size());
	}

	/**
	 * Implementation
	 */
	observer_adapter(){
		s2p_connection.on_new_pack = [this](size_t id, void* data, size_t size){
			abmt::blob blk(data,size);
			handle_data(id,blk);
		};
	}

	void set_connection(abmt::io::tcp_ptr tcp){
		s2p_connection.set_sink(tcp);
		s2p_connection.set_source(tcp);
		tcp->on_close = [this]{
			on_disconnect();
		};
	}

	virtual void on_disconnect(){
		s2p_connection.set_sink(io::sink());
		s2p_connection.set_source(io::source());
	}


	void handle_data(size_t id, abmt::blob& blk){
		cmd c = (cmd)id;
		switch(c){
		case cmd::model_online:
			on_model_online();
			break;
		case cmd::ack_signal_def:
			on_ack_signal_def(blk);
			break;
		case cmd::ack_parameter_def:
			on_ack_parameter_def(blk);
			break;
		case cmd::ack_set_daq_list:
			on_ack_set_daq_list();
			break;
		case cmd::ack_set_paq_list:
			on_ack_set_paq_list();
			break;
		case cmd::daq_data:
			on_daq_data(blk);
			break;
		case cmd::paq_data:
			on_paq_data(blk);
			break;
		default:
			return;
		}
	}

	virtual void send(cmd id, const void* data, uint32_t size){
		s2p_connection.send((uint8_t)id,data,size);
	}

	virtual void send(cmd id, abmt::blob& blk){
		s2p_connection.send((uint8_t)id,blk.data,blk.size);
	}

	virtual ~observer_adapter(){

	};

}; // observer_adapter


class model_adatper{
public:

	/**
	 * Interface
	 */

	abmt::io::s2p connection;
	abmt::rt::model* mdl = 0;
	abmt::mutex send_mtx;

	virtual void on_hello(uint16_t version){

	};

	virtual void on_request_signal_def(){

	};

	virtual void on_request_parameter_def(){

	};

	virtual void on_set_daq_list(abmt::blob& data){
		send(cmd::ack_set_daq_list,0,0);
	};

	virtual void on_set_paq_list(abmt::blob& data){
		send(cmd::ack_set_paq_list,0,0);
	};

	virtual void on_set_parameter(abmt::blob& data){

	};

	virtual void on_command(std::string command){

	};

	void send_model_online(){
		send(cmd::model_online,0,0);
	}

	/**
	 * Implementation
	 */

	model_adatper(){
		// empty...
		connection.on_new_pack = [this](size_t id, void* data, size_t size){
			abmt::blob blk(data,size);
			handle_data(id,blk);
		};
	}

	void set_connection(abmt::io::tcp_ptr tcp){
		connection.set_sink(tcp);
		connection.set_source(tcp);
		tcp->on_close = [this]{
			connection.set_sink(io::sink());
			connection.set_source(io::source());
		};
	}

	void set_model(abmt::rt::model* m){
		mdl = m;
	}

	void handle_data(size_t id, abmt::blob& blk){
		cmd c = (cmd)id;
		switch(c){
		case cmd::hello:
			on_hello(blk.get<uint16_t>(0));
			break;
		case cmd::request_signal_def:
			on_request_signal_def();
			break;
		case cmd::request_parameter_def:
			on_request_parameter_def();
			break;
		case cmd::set_daq_list:
			on_set_daq_list(blk);
			break;
		case cmd::set_paq_list:
			on_set_paq_list(blk);
			break;
		case cmd::set_prameter:
			on_set_parameter(blk);
			break;
		case cmd::command:
			on_command(blk.str());
			break;
		default:
			return;
		}
	}

	virtual void send(cmd id, const void* data = 0, uint32_t size = 0){
		auto lock = send_mtx.get_scope_lock();
		connection.send((uint8_t)id,data,size);
	}

	virtual void send(cmd id, abmt::blob& blk){
		auto lock = send_mtx.get_scope_lock();
		connection.send((uint8_t)id,blk.data,blk.size);
	}

	virtual ~model_adatper(){

	};

}; // class model_adapter


} // namespace rt
} // namespace abmt


#endif /* RUNTIME_INTERFACE_MODEL_COM_H_ */
