/**
 * Author: Hendrik van Arragon, 2021
 * SPDX-License-Identifier: MIT
 */

#include <abmt/serialize.h>
#include <cstring>
#include <cmath>
#include <iostream>
using namespace abmt::serialize;
using namespace std;

// helper:
// for i in bool char uint8_t int8_t uint16_t int16_t uint32_t int32_t uint64_t int64_t float double; do echo "type::type(${i}& o){"; echo -e "\ttype = type_t::${i^^};" | sed -e "s/_T//"; echo -e "\tobj = &o;"; echo -e "\tinfo.size = sizeof(o);"; echo -e "}"; echo ""; done
//

type::type(){
	id = type_id::UNDEFINED;
	obj = 0;
	serialize_info.size = 0;
}

type::type(bool& o){
	id = type_id::BOOL;
	obj = &o;
	serialize_info.size = sizeof(o);
}

type::type(char& o){
	id = type_id::CHAR;
	obj = &o;
	serialize_info.size = sizeof(o);
}

type::type(uint8_t& o){
	id = type_id::UINT8;
	obj = &o;
	serialize_info.size = sizeof(o);
}

type::type(int8_t& o){
	id = type_id::INT8;
	obj = &o;
	serialize_info.size = sizeof(o);
}

type::type(uint16_t& o){
	id = type_id::UINT16;
	obj = &o;
	serialize_info.size = sizeof(o);
}

type::type(int16_t& o){
	id = type_id::INT16;
	obj = &o;
	serialize_info.size = sizeof(o);
}

type::type(uint32_t& o){
	id = type_id::UINT32;
	obj = &o;
	serialize_info.size = sizeof(o);
}

type::type(int32_t& o){
	id = type_id::INT32;
	obj = &o;
	serialize_info.size = sizeof(o);
}

type::type(uint64_t& o){
	id = type_id::UINT64;
	obj = &o;
	serialize_info.size = sizeof(o);
}

type::type(int64_t& o){
	id = type_id::INT64;
	obj = &o;
	serialize_info.size = sizeof(o);
}

type::type(float& o){
	id = type_id::FLOAT;
	obj = &o;
	serialize_info.size = sizeof(o);
}

type::type(double& o){
	id = type_id::DOUBLE;
	obj = &o;
	serialize_info.size = sizeof(o);
}

type::type(std::string& o){
	id = type_id::STRING;
	obj = &o;
	serialize_info.rexp = {
			[](void* obj) -> size_t { return (*(std::string*)obj).size(); },
			[](void* obj, void* data, size_t size) { (*(std::string*)obj).copy((char*)data,size); },
			[](void* obj, void* data, size_t size) { (*(std::string*)obj) = std::string((char*)data,size); }
	};
}

type::type(member_list lst){
	id = type_id::OBJECT;
	obj = 0;
	serialize_info.attr = lst;
}
type::type(abmt::blob& blk){
	id = type_id::RAW_DATA;
	obj = 0;
	serialize_info.rexp = {
			[](void* obj) -> size_t { return (*(abmt::blob*)obj).size; },
			[](void* obj, void* data, size_t size) { (*(abmt::blob*)obj).get_mem(data, size, 0); },
			[](void* obj, void* data, size_t size) { (*(abmt::blob*)obj).set_mem(data, size, 0); }
	};
}

type::type(abmt::blob_shared& blk){
	id = type_id::RAW_DATA;
	obj = 0;
	serialize_info.rexp = {
			[](void* obj) -> size_t { return (*(abmt::blob*)obj).size; },
			[](void* obj, void* data, size_t size) { (*(abmt::blob*)obj).get_mem(data, size, 0); },
			[](void* obj, void* data, size_t size) { (*(abmt::blob*)obj).set_mem(data, size, 0); }
	};
}

type::type(blob_type r){
	id = type_id::RAW_DATA;
	obj = 0;
	serialize_info.rexp = r;
}

type::type(string_type s){
	id = type_id::STRING;
	obj = 0;
	serialize_info.rexp = s;
}

type::type(scene_2d_type s){
	id = type_id::SCENE_2D;
	obj = 0;
	serialize_info.rexp = s;
}

type::type(scene_3d_type s){
	id = type_id::SCENE_3D;
	obj = 0;
	serialize_info.rexp = s;
}

bool abmt::serialize::type::is_value(){
	if( id > type_id::UNDEFINED && id < type_id::OBJECT ){
		return true;
	}
	return false;
}
bool abmt::serialize::type::is_num(){
	if( id >= type_id::UINT8 && id < type_id::DOUBLE ){
		return true;
	}
	return false;
}
bool abmt::serialize::type::is_obj(){
	if( id == type_id::OBJECT || id == type_id::MAN_CTRL ){
		return true;
	}
	return false;
}

size_t type::get_size(){
	if( is_value() ){
		return serialize_info.size;
	}else if( is_obj() ){
		size_t res = 0;
		for(auto m:serialize_info.attr){
			auto& member_type = m.t;
			member_type.obj = ((char*)obj + m.offset);
			res += member_type.get_size();
		}
		return res;
	}else if( id == type_id::RAW_DATA || id == type_id::STRING || id == type_id::SCENE_2D || id == type_id::SCENE_3D){
		return serialize_info.rexp.get_size(obj) + 4; // 4 = len_info (uint32_t)
	}
	// todo: add other types!
	return 0;
}

double abmt::serialize::type::to_double(void* data_ptr, size_t data_len){
	double res = std::nan("");
	if(id == type_id::UINT8) {
		if(sizeof(uint8_t) > data_len) return res;
		res = (double) *(uint8_t*)data_ptr;
	} else if(id == type_id::INT8) {
		if(sizeof(int8_t) > data_len) return res;
		res = (double) *(int8_t*)data_ptr;
	} else if(id == type_id::UINT16) {
		if(sizeof(uint16_t) > data_len) return res;
		res = (double) *(uint16_t*)data_ptr;
	} else if(id == type_id::INT16) {
		if(sizeof(int16_t) > data_len) return res;
		res = (double) *(int16_t*)data_ptr;
	} else if(id == type_id::UINT32) {
		if(sizeof(uint32_t) > data_len) return res;
		res = (double) *(uint32_t*)data_ptr;
	} else if(id == type_id::INT32) {
		if(sizeof(int32_t) > data_len) return res;
		res = (double) *(int32_t*)data_ptr;
	} else if(id == type_id::UINT64) {
		if(sizeof(uint64_t) > data_len) return res;
		res = (double) *(uint64_t*)data_ptr;
	} else if(id == type_id::INT64) {
		if(sizeof(int64_t) > data_len) return res;
		res = (double) *(int64_t*)data_ptr;
	} else if(id == type_id::FLOAT) {
		if(sizeof(float) > data_len) return res;
		res = (double) *(float*)data_ptr;
	} else if(id == type_id::DOUBLE) {
		if(sizeof(double) > data_len) return res;
		res = (double) *(double*)data_ptr;
	}
	return res;
}

const member* member_list::begin(){
	return list;
}

const member* member_list::end(){
	return list + length;
}

scene_2d_type::scene_2d_type(blob_type bt):blob_type(bt){

}

scene_3d_type::scene_3d_type(blob_type bt):blob_type(bt){

}

size_t type::get_def_size(){
	size_t res = sizeof(type_id);
	if(is_obj()){
		res += 2; // name_len = 0 + OBJECT_END
		for(auto m:serialize_info.attr){
			res += 1; // 1 = name_len_byte
			auto name_len = strlen(m.name);
			if(name_len > 255){
				name_len = 255;
			}
			res += name_len;
			res += m.t.get_def_size();
		}
	}
	return res;
}

size_t type::get_def(void* def_buff, size_t def_size){
	if(def_size == 0) return 0;
	uint8_t* c_buf = (uint8_t*)def_buff;
	*c_buf = (uint8_t) id;
	size_t bytes_written_to_def = 1;
	if(is_obj()){
		if(def_size < 1){
			// not enough space for OBJECT_END
			*c_buf = (uint8_t) type_id::UNDEFINED;
			return 1;
		}
		for(auto m:serialize_info.attr){
			uint8_t* len_byte = c_buf + bytes_written_to_def;
			auto name_len = strlen(m.name);
			if(name_len > 255){
				name_len = 255;
			}

			// first 2 => len_byte + type_byte; second 2 => must be left over for name_len=0 and OBJECT_END
			if( (def_size - 2 - name_len) < 2 ){
				// not enough space for name
				*c_buf = (uint8_t) type_id::UNDEFINED;
				return 1;
			}

			*len_byte = (uint8_t) name_len;
			bytes_written_to_def += 1;
			uint8_t* name_str = c_buf + bytes_written_to_def;
			memcpy(name_str, m.name, name_len);
			bytes_written_to_def += name_len;

			uint8_t* member_type_pos = c_buf    + bytes_written_to_def;
			size_t   new_def_size    = def_size - bytes_written_to_def;
			size_t   type_size       = m.t.get_def(member_type_pos, new_def_size);
			bytes_written_to_def += type_size;
		}

		uint8_t* name_len_zero = c_buf + bytes_written_to_def;
		*name_len_zero = 0;
		uint8_t* obj_end = c_buf + bytes_written_to_def +1;
		*obj_end = (uint8_t) type_id::OBJECT_END;
		bytes_written_to_def += 2;
	}
	return bytes_written_to_def;
}

bool type::serialize(void* data, size_t data_size){
	return copy(data,data_size, true);
}

bool type::deserialize(void* data, size_t data_size){
	return copy(data,data_size);
}

bool deserialize(void* data, size_t data_size, void* def_buff, size_t def_size){
	/**
	 * Implementierungsidee:
	 * Wie to_json:
	 * Def und data paralell bis zum Ende laufen. Paralell dazu den eigenen Extractor
	 * durch iterrieren und copy aufrufen...
	 */
	return false;
}


size_t type::copy(void* data, size_t data_size, bool serialize){
	char* data_ptr = (char*) data;
	if( is_value() ){
		size_t size_needed = serialize_info.size;
		if(data_size >= size_needed){
			void* dst = data_ptr;
			void* src = obj;
			if(serialize == false){
				auto tmp = src;
				src = dst;
				dst = tmp;
			}
			memcpy(dst,src,size_needed);
			return size_needed;
		}
	}else if( is_obj() ){
		size_t data_pos = 0;
		char* obj_byte_ptr = (char*)obj;
		for(auto m:serialize_info.attr){
			// copy m.exp is important for threadsafty; m.exp does not know the obj, and you can not write it because
			// another obj might be serialized at the same time. type-copy-constructor does not work :(
			auto member_type = m.t;
			member_type.obj = (obj_byte_ptr + m.offset);
			size_t bytes_copied = member_type.copy(data_ptr + data_pos, data_size - data_pos, serialize);
			data_pos += bytes_copied;
		}
		return data_pos;
	}else if( id == type_id::RAW_DATA || id == type_id::STRING || id == type_id::SCENE_2D || id == type_id::SCENE_3D){
		if(serialize){
			uint32_t s = serialize_info.rexp.get_size(obj);
			if(data_size >= s + 4){
				*(uint32_t*)data_ptr = s;
				serialize_info.rexp.obj_to_data(obj,data_ptr + 4, data_size - 4);
				return s + 4;
			}
		}else{
			if( data_size < 4 ) return false;
			uint32_t s = *(uint32_t*) data_ptr;
			if( data_size >= s ){
				serialize_info.rexp.data_to_obj(obj,data_ptr + 4, s);
				return s + 4;
			}
		}
	}
	// todo: add other types!
	return 0;
}

